#include "game_of_life.h"
#include <stdlib.h>
#include <stdio.h>

#define LIVE 'o'
#define DEAD ' '

static unsigned char suma(unsigned char a, unsigned char b) {

     unsigned char r = 0x0;
     unsigned char mask = 0x1;

     while(mask){
          r |= (a&mask)^(b&mask)^
               ((((a&mask>>1)&(b&mask>>1))
                |((~r&mask>>1)&
                  ((a&mask>>1)|(b&mask>>1))))<<1);
          mask <<= 1;
     }
     return r;
}

/* 
 * Entrega un nuevo estado correspondiente a iterar el estado
 * anterior con la regla 23/3 
 */
ESTADO *step(ESTADO *e){

     int i, j;
     unsigned char suma_vecinas, new_cell;
     /* Nuevo estado*/
     ESTADO *new_e = new_estado(e->m, e->n);

     /* 
      * Setear bit (i,j) segun la siguiente regla 23/3 (Conway):
      * 
      *      - 1 si antes era 1 y esta rodeada por 2 ó 3 1's
      *        (si hay muchas vivas la célula muere por sobre-
      *        población, y si hay muy pocas muere por soledad)
      *      - 1 si antes era 0 y habían exactamente 3 vecinas
      *        en 1 (nace una nueva célula).
      *      - 0 si no se cumplió ninguna de las condiciones
      *        anteriores.
      */
     //printf("i < %d\n", 8*sizeof(unsigned char)*e->m);
     for(i = 0; i < 8*sizeof(unsigned char)*e->m; i++){
          for(j = 0; j < e->n; j++){
               suma_vecinas = suma(is_alive(e, i-1, j),
                                   suma(is_alive(e, i-1, j+1),
                                        suma(is_alive(e, i, j+1),
                                             suma(is_alive(e, i+1, j+1),
                                                  suma(is_alive(e, i+1, j),
                                                       suma(is_alive(e, i+1, j-1),
                                                            suma(is_alive(e, i, j-1),
                                                                 is_alive(e, i-1, j-1))))))));

	       #ifdef DEBUG
               printf("is_alive(e, %d,%d) = 0x%x\n", i, j, is_alive(e, i, j));
               printf("suma_vecinas = %d\n", suma_vecinas);
	       printf("regla_live = 0x%x \n", is_alive(e, i, j)&(!(suma_vecinas^0x02)|!(suma_vecinas^0x03)));
	       printf("suma_vecinas == 2? = 0x%x suma_vecinas == 3? = 0x%x\n", !(suma_vecinas^0x02), !(suma_vecinas^0x03));
	       printf("regla_dead = 0x%x \n", ~is_alive(e, i, j)&0x01&!(suma_vecinas^0x03));
	       #endif

	       new_cell = (is_alive(e, i, j)&(!(suma_vecinas^0x02)|!(suma_vecinas^0x03)))
	                  |
		          (~is_alive(e, i, j)&0x01&!(suma_vecinas^0x03));
               new_e->tablero[j][i>>3] |= new_cell<<(i%8);

	       #ifdef DEBUG
	       printf("new_cell = 0x%x\n", new_cell);
               printf("new_e->tablero[%d][%d>>3] = 0x%x\n", j, i, new_e->tablero[j][i>>3]);
	       #endif
                                          
          }
     }
     return new_e;

}

ESTADO *new_estado(int m, int n){

     int j;
     ESTADO *new_e = (ESTADO *)malloc(sizeof(ESTADO));
     new_e->tablero = (unsigned char **)malloc(sizeof(unsigned char *)*n);
     for(j = 0; j < n; j++){
        /* La funciÃ³n calloc es como malloc, pero se pide un arreglo de 1Âºarg
	 * eltos, donde cada elemento es de tamaÃ±lo 2Âºarg y la memoria viene en 0 */
        new_e->tablero[j] = (unsigned char *)calloc(m, sizeof(unsigned char));
     }
     new_e->m = m;
     new_e->n = n;
     return new_e;

}

void free_estado(ESTADO *e){

     int i, j;
     for(j = 0; j < e->n; j++)
         free(e->tablero[j]);
     free(e->tablero);
     free(e);

}

void print_estado(ESTADO *e){
   
     int i, j;

     #ifndef DEBUG
     system("clear");
     #endif

     for(j = 0; j < e->n; j++){
          for(i = 0; i < e->m; i++){
               print_char(e->tablero[j][i]);
          }
          printf("\n");
    }

}

static void print_char(unsigned char c){

     unsigned char mask = 0x01;
     while(mask){
          if(c&mask)
               putchar(LIVE);
          else
               putchar(DEAD);
	  putchar(' ');
          mask <<= 1;
     }

}

static unsigned char is_alive(ESTADO *e, int i, int j){

     if(i < 0 || j < 0 ||
        i >= 8*e->m*sizeof(unsigned char) || j >= e->n)
          return 0x00;
     return (e->tablero[j][i>>3]&(0x01<<(i%8)))>>(i%8);

}

int main(int argc, char **argv){
 
     ESTADO *e;
     int m, n, i, j;
     if(argc < 3){
          printf("Use game_of_life m n\n");
	  return EXIT_FAILURE;
     }
     m = atoi(argv[1]);
     n = atoi(argv[2]);
     e = new_estado(m, n);
     for(j = 0; j < n; j++)
          for(i = 0; i < m; i++)
               e->tablero[j][i] = rand()%256;

     #ifdef DEBUG
     printf("--------------\n");
     #endif

     ESTADO *f;
     while(1){
          print_estado(e);
          f = step(e);
          free_estado(e);
          e = f;

          getchar();
	  //sleep(1);
     }
     return EXIT_SUCCESS;

}
